import os
import shutil
from os import path as _p
from unittest import TestCase

import sh

from ..gitutil import git_expand
from tempfile import TemporaryDirectory


class TestGit(TestCase):
    tmpdir: TemporaryDirectory
    test_repo: str

    def setUp(self):
        self.tmpdir = TemporaryDirectory()
        orig_test_repo = _p.join(_p.dirname(__file__), 'hello-git-repo')
        self.test_repo = _p.join(self.tmpdir.name, 'hello-git-repo')
        shutil.copytree(orig_test_repo, self.test_repo)
        os.rename(_p.join(self.test_repo, '.gitted'), _p.join(self.test_repo, '.git'))

    def tearDown(self) -> None:
        self.tmpdir.cleanup()

    def test_expand(self):
        test_repo_bkup = _p.join(self.tmpdir.name, 'hello-git-repo.bkup')

        sh.cp(self.test_repo, test_repo_bkup, r=True)

        git_expand(self.test_repo, self.tmpdir.name, 'branch-A', 'branch-B', 'branch-C')

        with open(_p.join(self.tmpdir.name, 'branch-A', 'A'), 'r') as f:
            assert f.read() == 'A'

        with open(_p.join(self.tmpdir.name, 'branch-B', 'B'), 'r') as f:
            assert f.read() == 'B'

        with open(_p.join(self.tmpdir.name, 'branch-C', 'C'), 'r') as f:
            assert f.read() == 'C'

        assert not str(sh.diff(self.test_repo, test_repo_bkup, r=True))



from unittest import TestCase

from yautil import MpUtil


def double(x: int) -> int:
    return x * 2


class Object:
    i: int
    s: str
    d: dict
    l: list


def new_obj() -> Object:
    o = Object()
    o.i = 0
    o.s = 'object'
    o.d = {'a': 1, 'b': 2}
    o.l = ['c', 'd']

    return o


class TestMpUtil(TestCase):

    def test_basic(self):
        data = [0, 1, 2, 3, 4, 5]

        with MpUtil(total=len(data), pbar=False) as mp:
            for d in data:
                mp.schedule(double, d)
            results = mp.wait()

        for y in [0, 2, 4, 6, 8, 10]:
            assert y in results

    # def test_ordered(self):
    #     data = [0, 1, 2, 3, 4, 5]
    #
    #     with MpUtil(total=len(data), ordered=True) as mp:
    #         for d in data:
    #             mp.schedule(double, d)
    #         results = mp.wait()
    #
    #     for i, y in enumerate([0, 2, 4, 6, 8, 10]):
    #         assert y == results[i]

    def test_call_twice(self):
        self.test_basic()
        self.test_basic()

    def test_obj_ret(self):
        total = 10
        with MpUtil(total=total, pbar=False) as mp:
            for i in range(total):
                mp.schedule(new_obj)
            results = mp.wait()
        assert len(results) == total
        for o in results:
            assert isinstance(o, Object)
            assert o.i == 0
            assert o.s == 'object'
            assert o.d['a'] == 1
            assert o.d['b'] == 2
            assert len(o.l) == 2
            assert o.l[0] == 'c'
            assert o.l[1] == 'd'




import os
import shutil
import subprocess
import sys
import tempfile
from os import path as _p
from typing import Union, List

import sh

from .decorators import static_vars


def remove_contents(folder: str):
    for filename in os.listdir(folder):
        file_path = os.path.join(folder, filename)
        try:
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)
        except Exception as e:
            print('Failed to delete %s. Reason: %s' % (file_path, e))


def find_recursive_unix(root: str, name_patterns: list =None, ignored_dirs=None, type='any', depth=-1, sort=False):
    # find . -type d \( -path dir1 -o -path dir2 -o -path dir3 \) -prune -o -print
    find_cmd = 'find ' + root + ' '
    find_filter = ''

    if depth >= 0:
        find_cmd += '-maxdepth ' + str(depth) + ' '

    if name_patterns:
        find_filter += r'\( '
        find_filter += '-o '.join('-name \'' + e + '\' ' for e in name_patterns)
        find_filter += r'\) '

    if type == 'dir':
        find_filter += '-type d '
    elif type == 'file':
        find_filter += '-type f '
    elif type == 'any':
        pass
    else:
        return

    if ignored_dirs:
        find_cmd += r'-type d \( '
        while ignored_dirs:
            find_cmd += '-path ' + os.path.join(root, ignored_dirs.pop()) + ' '
            if ignored_dirs:
                find_cmd += '-o '
        find_cmd += r'\) -prune -o ' + find_filter + '-print'
    else:
        find_cmd += find_filter
    bout = subprocess.check_output(find_cmd, shell=True)
    out = bout.decode('utf-8').splitlines()
    if sort:
        out.sort()
    return out


def check_ignored(path, ignored_dirs):
    path = os.path.normpath(path)
    for dir in ignored_dirs:
        if path == dir or path.startswith(dir + os.path.sep):
            return True
    return False


def __find(root: str,
           name: Union[str, List[str]] = None,
           type: str = None,
           depth: int = None,
           exclude_dir: Union[str, List[str]] = None,
           iter_nonblock: bool = False
           ) -> any:

    find_opts = [root]
    find_target_opts = []

    if depth is not None:
        find_opts.extend(['-maxdepth', depth])

    if name:
        if not isinstance(name, list):
            name = [name]

        find_target_opts += '('
        while name:
            find_target_opts.extend(['-name', name.pop()])
            if name:
                find_target_opts.append('-o')
        find_target_opts += ')'

    if type:
        find_target_opts.extend(['-type', type])

    if exclude_dir:
        if not isinstance(exclude_dir, list):
            exclude_dir = [exclude_dir]

        find_opts.extend(['-type', 'd', '('])
        while exclude_dir:
            find_opts.extend(['-path', _p.join(root, exclude_dir.pop())])
            if exclude_dir:
                find_opts.append('-o')
        find_opts.extend([')', '-prune', '-o', *find_target_opts, '-print'])
    else:
        find_opts.extend(find_target_opts)

    # print(find_opts)

    for path in sh.find(*find_opts, _iter_noblock=iter_nonblock, _iter=True):
        if path:
            yield str(path).strip()


def find(root: str,
         name: str = None,
         type: str = None,
         depth: int = None,
         exclude_dir: Union[str, List[str]] = None,
         iter: bool = False,
         iter_nonblock: bool = False,
         ) -> any:
    ret = __find(root=root, name=name, type=type, depth=depth, exclude_dir=exclude_dir,
                 iter_nonblock=iter_nonblock)
    if iter or iter_nonblock:
        return ret
    else:
        return [*ret]


def find_recursive(root: str, name_patterns: list = None, ignored_dirs=None, type='any', depth=-1, sort=False):
    if sys.platform == "darwin" or sys.platform.startswith('linux'):
        # return find_recursive_unix(root, name_patterns=name_patterns, ignored_dirs=ignored_dirs, type=type, depth=depth, sort=sort)
        paths = find(root,
                     name=name_patterns,
                     type='d' if type == 'dir' else 'f' if type == 'file' else None,
                     depth=depth if depth >= 0 else None,
                     exclude_dir=ignored_dirs)
        return paths.sort() if sort else paths
    else:
        paths = []
        for root, dirs, fnames in os.walk(root):
            if ignored_dirs and check_ignored(root, ignored_dirs):
                continue
            if type == 'any' or type == 'dir':
                for dir in dirs:
                    if ignored_dirs and check_ignored(dir, ignored_dirs):
                        continue
                    paths.append(os.path.join(root, dir))
            if type == 'any' or type == 'file':
                for fname in fnames:
                    paths.append(os.path.join(root, fname))
        return paths


@static_vars(mkdir=sh.mkdir.bake(p=True),
             rsync=sh.rsync.bake(a=True, partial=True, delete=True))
def overwrite(src: str, dst: str):
    if os.path.isdir(src):
        src = src + os.path.sep

    if not os.path.exists(dst):
        overwrite.mkdir(os.path.dirname(dst))
    elif os.path.isdir(dst):
        dst = dst + os.path.sep

    overwrite.rsync(src, dst)


def get_memtmpdir(suffix=None, prefix=None, dir=None):
    if dir:
        return tempfile.TemporaryDirectory(suffix=suffix, prefix=prefix, dir=dir)

    memdirs = ['/dev/shm']
    for dir in memdirs:
        if not os.path.isdir(dir):
            continue
        return tempfile.TemporaryDirectory(suffix=suffix, prefix=prefix, dir=dir)

    return None





# https://stackoverflow.com/a/279586/3836385
def static_vars(**kwargs):
    def decorate(func):
        for k in kwargs:
            setattr(func, k, kwargs[k])
        return func
    return decorate



from typing import Union


class Event:
    def on_event(self) -> None:
        raise NotImplementedError

    @classmethod
    def gen_event(cls, base_event):
        raise NotImplementedError


class EventGenerator:
    events: dict

    def __init__(self):
        self.events = {}

    def register_event(self, base_event_cls, derived_event_cls):
        if base_event_cls in self.events:
            self.events[base_event_cls].append(derived_event_cls)
        else:
            self.events[base_event_cls] = [derived_event_cls]

    def throw_event(self, event):
        event.on_event()
        try:
            for derived_event_class in self.events[event.__class__]:
                derived_event = derived_event_class.gen_event(event)
                if derived_event is not None:
                    self.throw_event(derived_event)
        except KeyError:
            pass



import os
from os import path as _p
from unittest import TestCase

from yautil import docker_sh
from tempfile import TemporaryDirectory


class TestDocker(TestCase):
    tmpdir: TemporaryDirectory

    def setUp(self):
        self.tmpdir = TemporaryDirectory()

    def tearDown(self) -> None:
        self.tmpdir.cleanup()

    # def test_hello(self):
    #     uid = os.getuid()
    #     gid = os.getgid()
    #
    #     ctx = _p.join(self.tmpdir.name, 'test_hello')
    #     os.makedirs(ctx, exist_ok=True)
    #
    #     dockerfile = _p.join(ctx, 'Dockerfile')
    #     with open(dockerfile, 'w+') as f:
    #         f.write('FROM ubuntu')
    #
    #     assert str(dsh('echo', 'hello', _build_context=ctx)).strip() == 'hello'
    #
    #     assert int(str(dsh('id -u', _build_context=ctx)).strip()) == uid
    #
    #     assert int(str(dsh('id -g', _build_context=ctx)).strip()) == gid
    #
    #     assert int(str(dsh('id -u', _build_context=ctx, _root=True)).strip()) == '0'
    #
    #     assert int(str(dsh('id -g', _build_context=ctx, _root=True)).strip()) == '0'

    def test_docker_sh(self):
        uid = os.getuid()
        gid = os.getgid()

        ctx = _p.join(self.tmpdir.name, 'test_docker_sh')
        os.makedirs(ctx, exist_ok=True)

        dockerfile = _p.join(ctx, 'Dockerfile')
        with open(dockerfile, 'w+') as f:
            f.write('FROM ubuntu')

        c = docker_sh(ctx, root=True)

        assert str(c.echo('-n', 'hello')) == 'hello'

        assert int(str(c.id(u=True)).strip()) == 0

        assert int(str(c.id(g=True)).strip()) == 0

    def test_priv_drop(self):
        uid = os.getuid()
        gid = os.getgid()

        ctx = _p.join(self.tmpdir.name, 'test_priv_drop')
        os.makedirs(ctx, exist_ok=True)

        dockerfile = _p.join(ctx, 'Dockerfile')
        with open(dockerfile, 'w+') as f:
            f.write('FROM ubuntu')

        c = docker_sh(ctx)

        assert str(c.echo('-n', 'hello')) == 'hello'

        assert int(str(c.id(u=True)).strip()) == uid

        assert int(str(c.id(g=True)).strip()) == gid

    def test_verbose(self):
        uid = os.getuid()
        gid = os.getgid()

        ctx = _p.join(self.tmpdir.name, 'test_verbose')
        os.makedirs(ctx, exist_ok=True)

        dockerfile = _p.join(ctx, 'Dockerfile')
        with open(dockerfile, 'w+') as f:
            f.write('FROM ubuntu')

        c = docker_sh(ctx, verbose=True)

        assert str(c.echo('-n', 'hello')) == 'hello'

        assert int(str(c.id(u=True)).strip()) == uid

        assert int(str(c.id(g=True)).strip()) == gid
